#!/usr/local/bin/perl -w
require 5.003;
use strict;

#
# takes a special file name pattern as an input (see below)
# searches for corresponding files
# groups files by Run parameters
# for each Run, runs 'lx' for all thread logs to collect avg stats
# and builds a summary table with stats for each Run as rows
#
# input pattern:
# 	whatever:RUN:whatever
# where 
#	'whatever' stands for any characters or wild cards in a file name
#	':'        stands for itself, i.e., character ':'
# 	'RUN'      regular expression, e.g., '*' character
#                  ':'s will be removed during the search, and the 
#	           substrings matching RUN will be taken as
#	           Run parameters for grouping purposes (see above)
#
# there must be at least one input pattern
# there must be at least one :RUN: entry per pattern
#
# Example:
#
# ./exprep.pl 'logs/bakeoff12/clt.10.0.1.??.repeat5.rr:*:.log'

#

my (@Patterns) = @ARGV;
die("usage: $0 <fname_pattern> ...\n") unless @Patterns;

# Configuration

# location of the 'lx' program
my $LogExtracter = './lx';

# objects to include in the summary table
my @Objects = qw(
	req_rate rep_rate req_dhr rep_dhr 
	resp_tm_mean
	resp_tm_perc:25 resp_tm_perc:50 resp_tm_perc:75
	err_cnt err_pct
	log_cnt
	);

# globals

my %Runs = (); # run descriptions indexed by run names (parameters)
my @UserOrder = ();  # runs in the order they were found
my @SmartOrder = (); # number-names first; all in order

exit(&main());

sub main {

	# there must be at least one :RUN: entry per pattern
	my ($badPattern) = grep { !/:.+:/ } @Patterns;
	die("pattern '$badPattern' does not have a :RUN: card; stopped") 
		if defined $badPattern;

	# get run descriptions for each pattern
	foreach (@Patterns) { &getRuns($_); }
	die("no files matching input pattern(s); stopped") unless keys %Runs;

	@SmartOrder = sort cmpRuns @UserOrder;

	# report our findings
	&reportRuns();

	# get actual stats
	foreach (@UserOrder) { &getRunStats($_); }

	# build summary table
	&prnTable();

	return 0;
}

sub getRuns {
	my $pattern = shift;

	# replace RE with shell wild card (*)
	my $shell_pattern = shellPattern($pattern);

	# get all the files
	my @files = glob($shell_pattern);

	# filter out what does not match the perl pattern
	my $perl_pattern = perlPattern($pattern);
	@files = grep { /$perl_pattern/ } @files;

	if (!scalar @files) {
		warn("no files matched '$pattern', ignored");
		warn("shell pattern used: '$shell_pattern'\n");
		warn("perl pattern used:  '$perl_pattern'\n");
		return undef;
	}

	# find distinct runs
	foreach (@files) {
		my (@matches) = /$perl_pattern/g;
		die("internal error: '$perl_pattern' does not match '$_'; stopped")
			unless @matches;
		my $params = join('_', @matches);

		my $run = $Runs{$params};
		if (!$run) {
		 	$run = { 
				name => $params,
				pattern => $pattern,
				threads => [],
			};
			$Runs{$params} = $run;
			push @UserOrder, $run;
		}
		push @{$run->{threads}}, $_;
	}
}

sub reportRuns {

	foreach (@SmartOrder) {
		print("#\n");
		printf("# run: %s\n", $_->{name});
		printf("#  pattern: %s\n", $_->{pattern});
		printf("#   threads (%d):\n", scalar @{$_->{threads}});
		foreach my $th (@{$_->{threads}}) {
			printf("#    %s\n", $th);
		}
	}
	print("#\n");
}

sub getRunStats {
	my $run = shift;

	my $cmd = sprintf('%s --objects %s %s',
		$LogExtracter,
		join(',', @Objects),
		join(' ', @{$run->{threads}}));
		
	my $text = `$cmd`;
	die("failed: '$cmd', $!; stopped") unless $text;

	# remove comments if any
	$text =~ s/^\s*#.*$//gm;

	# trim
	$text =~ s/^\s*|\s*$//gm;

	$run->{stats} = $text;
}

sub prnTableRec {
	my $col_widths = shift;
	my @cws = @{$col_widths};

	printf('%-*s', shift @cws, shift @_);
	map { printf(' %*s', shift @cws, $_) } @_;
	print("\n");
}

sub prnTable {

	# get column widths

	my @col_widths = (4, (0) x scalar @Objects);

	my $col;
	for ($col = 1; $col <= $#col_widths; ++$col) {
		my $len = length $Objects[$col-1];
		$col_widths[$col] = $len if $len > $col_widths[$col]; 
	}

	foreach my $run (@SmartOrder) {
		my @flds =  ($run->{name}, split /\s+/, $run->{stats});
		for ($col = 0; $col <= $#col_widths; ++$col) {
			my $len = length $flds[$col];
			$col_widths[$col] = $len if $len > $col_widths[$col];
		}
	}
	
	# print header
	prnTableRec(\@col_widths, '#Run', @Objects);

	# print table body
	foreach my $run (@SmartOrder) {
		prnTableRec(\@col_widths, $run->{name},
			split /\s+/, $run->{stats});
	}
}

sub shellPattern {
	my $pattern = shift;

	# shell does not understand our :card:
	$pattern =~ s/:.*?:/*/g;

	return $pattern;
}

sub perlPattern {
	my $pattern = shift;

	# some people might use shell RE in our :card:
	$pattern =~ s/:\*:/:.*:/g;
	$pattern =~ s/:\?:/:.?:/g;

	# perl is not shell, we must substitute all shell RE and protect 
	# other chars outside :card:
	my @parts = split(/:/, $pattern);
	$pattern = '';
	my $state = 1;
	foreach (@parts) {
		if ($state > 0) {
			s/\./\\./g;
			s/\*/.*/g;
			s/\?/.?/g;
		}
		$pattern .= ':' if length $pattern;
		$pattern .= $_;

		$state = 1 - $state;
	}

	# perl does not understand our :card:
	$pattern =~ s/:(.*?):/($1)/g;

	return $pattern;
}

sub isInt {
	return shift =~ m/^[\-\+]?\d+(?:.?\d*)?$/;
}

sub cmpRuns {
	if (&isInt($a->{name}) && &isInt($b->{name})) {
		return $a->{name} <=> $b->{name};
	} else {
		return $a->{name} cmp $b->{name};
	}
}
